#!/usr/bin/env python

# After converting the .TIL file into a binary tree,
# make 3 passes through that data,
# converting class, method, and variable references.
# This will enable the Treenimation Runtime, which takes as input
# the binary tree and executes it, to do its job more efficiently.

import sys
import string
import tilLoader
try:
    import pygtk
    pygtk.require("2.0")
except:
    pass
try:
    import gtk
    import gtk.glade
    import os
    import pango
except:
    sys.exit(1)

class NtsPkg:  #gpk

    def __init__(self):
        self.pkgName = 0
        self.pkgDecl = 1
        self.clsInterfDecl = 2

class NtsClassDecl:  #gcd

    def __init__(self):
        self.cls = 0
        self.id = 1
        self.modifs = 2
        self.clsIdxPtr = 3
        self.interfRef = 4
        self.fldDecl = 5
        self.methDecl = 6
        self.staticBlk = 7
        self.doBlk = 8

class NtsMethDecl:  #gmd

    def __init__(self):
        self.methTyp = 0
        self.funcTyp = 1
        self.modifs = 2
        self.parmDecl = 3
        self.locDecl = 4
        self.exceptTyp = 5
        self.block = 6
        self.id = 7

class ConvClsVarMth:

    def __init__(self, listBook):
        self.book = listBook
        self.ntMask = 0x000F
        self.nt = tilLoader.NodeType()
        self.gpk = NtsPkg()
        self.gcd = NtsClassDecl()
        self.gmd = NtsMethDecl()
        self.debug = listBook.debug
        self.status = False
        self.suppressDebug = False
        if not self.convertClsIdxPtrs():   # 1st pass
            return
        self.suppressDebug = False
        if not self.convertClsRefPtrs():   # 2nd pass
            return
        self.suppressDebug = False
        if not self.convertVarMthIdxPtrs():   # 3rd pass
            return
        self.status = True

    def showDebugMsg(self, msg):
        if self.suppressDebug:
            return
        msg = msg + '\n\n' + 'Continue debug messages?'
        dialog = gtk.MessageDialog(None,
            #parent_window,  # the window that spawned this dialog
            gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
            gtk.MESSAGE_QUESTION,
            gtk.BUTTONS_NONE,
            msg)
        dialog.add_buttons(
            gtk.STOCK_YES, gtk.RESPONSE_YES,
            gtk.STOCK_NO, gtk.RESPONSE_NO,
            gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        dialog.show_all()
        result = dialog.run()
        if result == gtk.RESPONSE_YES:
            #print 'Yes'
            pass
        elif result == gtk.RESPONSE_NO:
            #print 'No'
            self.suppressDebug = True
        elif result == gtk.RESPONSE_CANCEL:
            #print 'Cancel'
            self.debug = False
            self.book.debug = False
        dialog.destroy()

    def convertClsIdxPtrs(self):
        pkgCount = 0
        if not self.book.goChild():
            return False
        while True:
            if not self.book.goChild():
                return False
            if not self.book.goNextSibling():
                return False
            if not self.book.goChild():
                return False
            pkgHdrIdx = 0
            while True:
                if not self.book.goChild():
                    return False
                hdrVal = self.book.currListNode.header & self.ntMask
                if hdrVal != self.nt.int:
                    return False
                pkgIdx = self.book.currListNode.left
                if self.debug:
                    self.showDebugMsg('pkgIdx = %d' % pkgIdx)
                if not self.book.goNextSibling():
                    return False
                hdrVal = self.book.currListNode.header & self.ntMask
                if hdrVal != self.nt.int:
                    return False
                clsIdx = self.book.currListNode.left                    
                if self.debug:
                    self.showDebugMsg('clsIdx = %d' % clsIdx)
                if not self.book.goParent():
                    return False
                if not self.book.goParent():
                    return False
                if not self.book.goParent():
                    return False
                if not self.book.goParent():
                    return False
                if not self.book.goChild():
                    return False
                for i in range(pkgIdx):
                    if not self.book.goNextSibling():
                        return False
                if not self.book.goChild():
                    return False
                if not self.doMultiRight(self.gpk.clsInterfDecl):
                    return False
                if not self.book.goChild():
                    return False
                for i in range(clsIdx):
                    if not self.book.goNextSibling():
                        return False
                clsPtr = self.book.currWordIdx
                if self.debug:
                    self.showDebugMsg('clsPtr = %d' % clsPtr)
                if not self.book.goParent():
                    return False
                if not self.book.goParent():
                    return False
                if not self.book.goParent():
                    return False
                if not self.book.goChild():
                    return False
                for i in range(pkgCount):
                    if not self.book.goNextSibling():
                        return False
                if not self.book.goChild():
                    return False
                if not self.book.goNextSibling():
                    return False
                if not self.book.goChild():
                    return False
                for i in range(pkgHdrIdx):
                    if not self.book.goNextSibling():
                        return False
                if not self.book.goChild():
                    return False
                if not self.book.goNextSibling():
                    return False
                hdrVal = self.book.currListNode.header & self.ntMask
                if hdrVal != self.nt.int:
                    return False
                self.book.currListNode.left = clsPtr
                self.book.setListNode(
                    self.book.currListNode, self.book.currWordIdx)
                if not self.book.goParent():
                    return False
                pkgHdrIdx = pkgHdrIdx + 1
                if not self.book.goNextSibling():
                    break
            if not self.book.goParent():
                return False
            if not self.book.goParent():
                return False
            pkgCount = pkgCount + 1                
            if not self.book.goNextSibling():
                break
        if not self.book.goParent():
            return False
        if self.debug:
            self.showDebugMsg('pkg count = %d' % pkgCount)
        return True

    def convertClsRefPtrs(self):
        hiLongInt = 0x7FFFFFFF  # 2^31 - 1
        listNode = tilLoader.ListNode(0, 0)
        nt = tilLoader.NodeType()
        book = self.book
        pkgCount = 0
        firstPkg = True
        if not book.goChild():
            return False
        pkgStartIdx = book.currWordIdx
        while True:
            lastPkg = not book.goNextSibling()
            if firstPkg and lastPkg:
                return False
            if firstPkg:
                pkgEndIdx = book.currWordIdx
                if not book.goPrevSibling():
                    return False
            elif lastPkg:
                pkgEndIdx = hiLongInt
                if not book.goPrevSibling():
                    return False
            elif not book.goNextSibling():
                lastPkg = True
                pkgEndIdx = hiLongInt
            else:
                pkgEndIdx = book.currWordIdx
                if not book.goPrevSibling():
                    return False
            maxWordIdx = book.makePtrVal(book.maxPageIdx, book.maxNodeIdx)
            wordIdx = pkgStartIdx
            while wordIdx <= pkgEndIdx:
                if wordIdx > maxWordIdx:
                    break
                book.getListNode(listNode, wordIdx)
                hdrVal = listNode.header & self.ntMask
                if hdrVal != nt.keyword:
                    wordIdx = book.getNextWordIdx(wordIdx)
                    continue
                i = listNode.left
                isClsLoc = (book.keyWordList[i] == 'cls-loc')
                if (not isClsLoc) and (book.keyWordList[i] != 'cls-glb'):
                    wordIdx = book.getNextWordIdx(wordIdx)
                    continue
                nextWordIdx = book.getNextWordIdx(wordIdx)
                if nextWordIdx >= maxWordIdx:
                    return False
                book.getListNode(listNode, nextWordIdx)
                hdrVal = listNode.header & self.ntMask
                if hdrVal != nt.int:
                    return False
                clsIdx = listNode.left
                if isClsLoc:
                    clsPtr = self.convertClsLocPtr(clsIdx)
                else:
                    clsPtr = self.convertClsGlbPtr(clsIdx)
                if clsPtr < 0:
                    return False
                listNode.left = clsPtr
                book.setListNode(listNode, nextWordIdx)
                wordIdx = book.getNextWordIdx(wordIdx)
            firstPkg = False
            pkgStartIdx = pkgEndIdx
            pkgCount = pkgCount + 1
            if lastPkg:
                break
        if not book.goParent():
            return False
        if self.debug:
            self.showDebugMsg('convertClsRefPtrs: pkg count = %d' % pkgCount)
        return True

    def convertClsLocPtr(self, clsIdx):
        result = -1
        book = self.book
        if not book.goChild():
            return result
        if not self.doMultiRight(self.gpk.clsInterfDecl):
            return result
        if not book.goChild():
            return result
        for i in range(clsIdx):
            if not book.goNextSibling():
                return result
        i = book.currWordIdx
        if not book.goParent():
            return result
        if not book.goParent():
            return result
        result = i
        if self.debug:
            msg = 'clsIdx = %d' % clsIdx
            msg = msg + (', clsPtr = %d' % result)
            self.showDebugMsg(msg)
        return result

    def convertClsGlbPtr(self, clsIdx):
        nt = tilLoader.NodeType()
        result = -1
        book = self.book
        if not book.goChild():
            return result
        if not book.goNextSibling():
            return result
        if not book.goChild():
            return result
        for i in range(clsIdx):
            if not book.goNextSibling():
                return result
        if not book.goChild():
            return result
        if not book.goNextSibling():
            return result
        hdrVal = book.currListNode.header & self.ntMask
        if hdrVal != nt.int:
            return result
        i = book.currListNode.left
        if not book.goParent():
            return result
        if not book.goParent():
            return result
        if not book.goParent():
            return result
        result = i
        if self.debug:
            msg = 'clsIdx = %d' % clsIdx
            msg = msg + (', clsPtr = %d' % result)
            self.showDebugMsg(msg)
        return result

    def convertVarMthIdxPtrs(self):
        book = self.book
        nt = tilLoader.NodeType()
        pkgIdx = 0
        if not book.goChild():
            return False
        while True:
            if not book.goChild():
                return False
            if not self.doMultiRight(self.gpk.clsInterfDecl):
                return False
            if not book.goChild():
                return False
            clsIdx = 0
            while True:
                if not book.goChild():
                    return False
                if not book.goNextSibling():
                    return False
                hdrVal = book.currListNode.header & self.ntMask
                if hdrVal != nt.identifier:
                    return False
                i = book.currListNode.left
                if self.debug:
                    self.showDebugMsg('cls name = ' + book.tilWordList[i])
                if not self.doRangeRight(self.gcd.id, self.gcd.methDecl):
                    return False
                if not book.goChild():
                    return False
                mthIdx = 0
                while True:
                    if not book.goChild():
                        return False
                    i = book.currListNode.left
                    if self.debug:
                        self.showDebugMsg('meth typ = ' + book.tilWordList[i])
                    if not self.doRangeRight(self.gmd.methTyp, self.gmd.block):
                        return False
                    if not self.scanMethConvertDeclIdx(pkgIdx, clsIdx, mthIdx):
                        return False
                    if not book.goParent():                    
                        return False
                    mthIdx = mthIdx + 1
                    if not book.goNextSibling():
                        break
                if not book.goParent():                    
                    return False
                if not book.goParent():                    
                    return False
                clsIdx = clsIdx + 1
                if not book.goNextSibling():
                    break
            if not book.goParent():                    
                return False
            if not book.goParent():                    
                return False
            pkgIdx = pkgIdx + 1
            if not book.goNextSibling():
                break
        if not book.goParent():                    
            return False
        if self.debug:
            self.showDebugMsg('convertVarMthIdxPtrs: pkg count = %d' % pkgIdx)
        return True

    def scanMethConvertDeclIdx(self, pkgIdx, clsIdx, mthIdx):
        listNode = tilLoader.ListNode(0, 0)
        tempListNode = tilLoader.ListNode(0, 0)
        nt = tilLoader.NodeType()
        book = self.book
        book.copyListNode(book.currListNode, tempListNode)
        startIdx = book.currWordIdx
        if not book.goNextSibling():
            return False
        endIdx = book.currWordIdx
        book.currWordIdx = startIdx
        while book.currWordIdx <= endIdx:
            book.getListNode(book.currListNode, book.currWordIdx)
            hdrVal = book.currListNode.header & self.ntMask
            if hdrVal != nt.keyword:
                book.currWordIdx = book.getNextWordIdx(book.currWordIdx)
                continue
            i = book.currListNode.left
            keyword = book.keyWordList[i]
            if not (
                (keyword == 'loc') or
                (keyword == 'parm') or
                (keyword == 'fld') or
                (keyword == 'meth')):
                book.currWordIdx = book.getNextWordIdx(book.currWordIdx)
                continue
            if not book.goNextSibling():
                return False
            isClsMid = not book.atomic(book.currListNode)
            if isClsMid:
                if not book.goChild():
                    return False
                hdrVal = book.currListNode.header & self.ntMask
                if hdrVal != nt.keyword:
                    return False
                i = book.currListNode.left
                isClsLoc = (book.keyWordList[i] == 'cls-loc')
                if not book.goNextSibling():
                    return False
                hdrVal = book.currListNode.header & self.ntMask
                if hdrVal != nt.int:
                    return False
                clsPtr = book.currListNode.left
                if not book.goParent():
                    return False
                if not book.goNextSibling():
                    return False
            else:
                isClsLoc = False
                clsPtr = 0
            hdrVal = book.currListNode.header & self.ntMask
            if hdrVal != nt.int:
                return False
            declIdx = book.currListNode.left
            book.copyListNode(book.currListNode, listNode)
            listNodeIdx = book.currWordIdx
            if self.debug:
                msg = 'keyword = ' + keyword
                msg = msg + (', decl idx = %d' % declIdx)
                if isClsMid:
                    msg = msg + ', ('
                    if isClsLoc:
                        msg = msg + '#cls-loc '
                    else:
                        msg = msg + '#cls-glb '
                    msg = msg + ('%d' % clsPtr)
                    msg = msg + ')'
                self.showDebugMsg(msg)
            savedWordIdx = book.currWordIdx
            if keyword == 'loc':
                declPtr = self.convertLocPtr(declIdx)
            elif keyword == 'parm':
                declPtr = self.convertParmPtr(declIdx)
            elif keyword == 'fld':
                if isClsMid:
                    declPtr = self.convertFldMthClsPtr(
                        False, clsPtr, declIdx, pkgIdx, clsIdx, mthIdx)
                else:
                    declPtr = self.convertFldPtr(mthIdx, declIdx)
            elif keyword == 'meth':
                if isClsMid:
                    declPtr = self.convertFldMthClsPtr(
                        True, clsPtr, declIdx, pkgIdx, clsIdx, mthIdx)
                else:
                    declPtr = self.convertMethPtr(mthIdx, declIdx)
            else:
                declPtr = -1
            if declPtr < 0:
                return False
            listNode.left = declPtr
            book.setListNode(listNode, listNodeIdx)
            # convert rtns lv curr node at parent of #proc/#func node
            if not book.goChild():
                return False
            if not self.doMultiRight(self.gmd.block):
                return False
            # back at body of method
            book.currWordIdx = savedWordIdx
            book.currWordIdx = book.getNextWordIdx(book.currWordIdx)
        book.copyListNode(tempListNode, book.currListNode)
        book.currWordIdx = startIdx
        return True

    def convertLocPtr(self, declIdx):
        result = -1
        book = self.book
        if not book.goParent():
            return result
        if not book.goChild():
            return result
        if not self.doMultiRight(self.gmd.locDecl):
            return result
        # at parent of 1st loc decl.
        if not book.goChild():
            return result
        for i in range(declIdx):
            if not book.goNextSibling():
                return result
        wordIdx = book.currWordIdx
        if not book.goParent():   # at parent of 1st loc decl.
            return result
        if not book.goParent():
            return result
        return wordIdx

    def convertParmPtr(self, declIdx):
        result = -1
        book = self.book
        if not book.goParent():
            return result
        if not book.goChild():
            return result
        if not self.doMultiRight(self.gmd.parmDecl):
            return result
        # at parent of 1st parm decl.
        if not book.goChild():
            return result
        for i in range(declIdx):
            if not book.goNextSibling():
                return result
        wordIdx = book.currWordIdx
        if not book.goParent():   # at parent of 1st parm decl.
            return result
        if not book.goParent():
            return result
        return wordIdx

    def convertFldPtr(self, mthIdx, declIdx):
        result = -1
        book = self.book
        if not book.goParent():
            return result
        if not book.goParent():
            return result
        if not book.goParent():
            return result
        if not book.goChild():
            return result
        if not self.doMultiRight(self.gcd.fldDecl):
            return result
        # at parent of 1st fld decl.
        if not book.goChild():
            return result
        for i in range(declIdx):
            if not book.goNextSibling():
                return result
        wordIdx = book.currWordIdx
        if not book.goParent():   # at parent of 1st fld decl.
            return result
        if not book.goNextSibling():
            return result
        if not book.goChild():
            return result
        for i in range(mthIdx):
            if not book.goNextSibling():
                return result
        return wordIdx

    def convertMethPtr(self, mthIdx, declIdx):
        result = -1
        book = self.book
        if not book.goParent():
            return result
        if not book.goParent():
            return result
        if not book.goChild():   # at parent of 1st #proc/#func node
            return result
        for i in range(declIdx):
            if not book.goNextSibling():
                return result
        wordIdx = book.currWordIdx
        if not book.goParent():   
            return result
        if not book.goChild():   # at parent of 1st #proc/#func node
            return result
        for i in range(mthIdx):
            if not book.goNextSibling():
                return result
        return wordIdx

    def convertFldMthClsPtr(self, isMeth, clsPtr, declIdx, pkgIdx, clsIdx, mthIdx):
        result = -1
        book = self.book
        book.currWordIdx = clsPtr
        book.getListNode(book.currListNode, book.currWordIdx)
        if not book.goChild():
            return result
        if isMeth:
            if not self.doMultiRight(self.gcd.methDecl):
                return result
        elif not self.doMultiRight(self.gcd.fldDecl):
            return result
        if not book.goChild():
            return result
        for i in range(declIdx):
            if not book.goNextSibling():
                return result
        wordIdx = book.currWordIdx
        if self.backToMethod(pkgIdx, clsIdx, mthIdx):
            result = wordIdx
        return result

    def backToMethod(self, pkgIdx, clsIdx, mthIdx):
        book = self.book
        book.resetStack()
        if not book.goChild():
            return False
        for i in range(pkgIdx):
            if not book.goNextSibling():
                return False
        if not book.goChild():
            return False
        if not self.doMultiRight(self.gpk.clsInterfDecl):
            return False
        if not book.goChild():
            return False
        for i in range(clsIdx):
            if not book.goNextSibling():
                return False
        if not book.goChild():
            return False
        if not self.doMultiRight(self.gcd.methDecl):
            return False
        if not book.goChild():
            return False
        for i in range(mthIdx):
            if not book.goNextSibling():
                return False
        return True

    def doRangeRight(self, startIdx, endIdx):
        return self.doMultiRight(endIdx - startIdx)

    def doMultiRight(self, count):
        for i in range(count):
            if not self.book.goNextSibling():
                return False
        return True
    
