#!/usr/bin/env python

# Main TIL Loader Python script
#
# 1. Enable user to select a .TIL file
# 2. TIL = Treenimation Intermediate Language
#    (output of the Treescript compiler: a text file)
# 3. Click on Load button to convert TIL file into a binary tree
# 4. Use arrow keys to traverse the binary tree in memory:
#    Left: go to parent node
#    Right: go to first child node
#    Down: go to next node
#    Up: go to previous node

import sys
import tilLoader
try:
    import pygtk
    pygtk.require("2.0")
except:
    pass
try:
    import gtk
    import gtk.glade
    import os
    import pango
except:
    sys.exit(1)

class FileSelectionEx:
    # Get the selected filename and print it to the console
    def file_ok_sel(self, w):
        fullFileName = self.filew.get_filename()
        print "%s" % fullFileName
        f = file(self.cfgFileName, 'w');
        try:
            f.write(fullFileName)
        finally:
            f.close()
        filePair = os.path.split(fullFileName)
        self.dirLbl.set_text(filePair[0])
        self.dirName = filePair[0]
        self.fileNameLbl.set_text(filePair[1])
        print "OK"
        self.destroy(w)

    def destroy(self, widget):
        self.filew.response(gtk.RESPONSE_CLOSE)
        self.filew.destroy()

    def __init__(self, cfgFileName, fileNameLbl, dirLbl):
        self.cfgFileName = cfgFileName
        self.fileNameLbl = fileNameLbl
        self.dirLbl = dirLbl
        self.dirName = ""
        self.fullFileName = os.path.join(
            dirLbl.get_text(), fileNameLbl.get_text())
        
        # Create a new file selection widget
        self.filew = gtk.FileSelection("File selection")
        self.filew.connect("destroy", self.destroy)

        # Connect the ok_button to file_ok_sel method
        self.filew.ok_button.connect("clicked", self.file_ok_sel)

        # Connect the cancel_button to destroy the widget
        self.filew.cancel_button.connect("clicked", self.destroy)
            #lambda w: self.filew.destroy())

        # Lets set the filename, as if this were a save dialog,
        # and we are giving a default filename
        self.filew.set_filename(self.fullFileName)
        self.filew.set_modal(True)

class HellowWorldGTK:
    """This is an Hello World GTK application"""
    global helloLabel
    global helloFlag

    def __init__(self):
        
        #Set the Glade file
        self.gladefile = "projTilLoader.glade"  
        self.wTree = gtk.glade.XML(self.gladefile)
        self.helloFlag = 1
        self.myBook = []

        #Create our dictionary and connect it
        dic = {
            "on_loadBtn_clicked" : self.loadBtn_clicked,
            "on_unusedBtn_clicked" : self.unusedBtn_clicked,
            "on_MainWindow_key" : self.keyPressed,
            "on_selectBtn_clicked" : self.selectBtn_clicked,
            "on_selectBtn_key_up" : self.selectBtn_clicked,
            "on_debugCb_toggled" : self.debugCb_toggled,
            "on_MainWindow_destroy" : gtk.main_quit }
        self.wTree.signal_autoconnect(dic)
        self.StartDisplay()
        self.SetBtnsAndLabels()

    def loadBtn_clicked(self, widget):
        fileNameLbl = self.wTree.get_widget("fileNameLabel")
        dirLbl = self.wTree.get_widget("dirLabel")
        if fileNameLbl.get_text() == "":
            return
        fullFileName = os.path.join(dirLbl.get_text(), fileNameLbl.get_text())
        print fullFileName
        textView = self.wTree.get_widget("tilTextView")
        buf = textView.get_buffer()
        f = file(fullFileName, 'r');
        try:
            fileContents = f.read()
        finally:
            f.close()
        buf.set_text(fileContents)
        tilCodeList = fileContents.splitlines()
        if not self.tilConstFiles.isConstFilesLoaded:
            return
        keyWordList = self.tilConstFiles.keyWordList
        enumList = self.tilConstFiles.enumList
        self.myBook = tilLoader.ListBook(
            tilCodeList, keyWordList, enumList, self.debug)
        myPage = self.myBook.pageBook[0]

    def unusedBtn_clicked(self, widget):
        self.helloFlag = not self.helloFlag
        self.SetBtnsAndLabels()
        self.wTree.get_widget("unusedBtn").set_label(self.helloLabel)

    def keyPressed(self, widget, event):
        keyname = gtk.gdk.keyval_name(event.keyval)
        #print "Key %s (%d) was pressed" % (keyname, event.keyval)
        oldCaption = self.wTree.get_widget("keyLabel").get_label()
        if keyname == 'Up':
            msg = self.upCmd(keyname)
        elif keyname == 'Down':
            msg = self.downCmd(keyname)
        elif keyname == 'Left':
            msg = self.leftCmd(keyname)
        elif keyname == 'Right':
            msg = self.rightCmd(keyname)
        else:
            msg = ''
        if msg != '':
            self.wTree.get_widget("keyLabel").set_label(msg)
        if msg == '':
            leftCaption = 'N/A'
        elif self.myBook == []:
            leftCaption = ''
        else:
            leftCaption = '%d' % self.myBook.currListNode.left
        if leftCaption != 'N/A':
            self.wTree.get_widget("leftLabel").set_label(leftCaption)

    def upCmd(self, keyname):
        return self.myBook.goPrevSiblingToken()

    def downCmd(self, keyname):
        return self.myBook.goNextSiblingToken()

    def rightCmd(self, keyname):
        return self.myBook.goChildToken()

    def leftCmd(self, keyname):
        return self.myBook.goParentToken()

    def selectBtn_clicked(self, widget):
        fileNameLbl = self.wTree.get_widget("fileNameLabel")
        dirLbl = self.wTree.get_widget("dirLabel")
        fileSel = FileSelectionEx(self.cfgFileName, fileNameLbl, dirLbl)
        dirStr = fileSel.dirName
        if dirStr != "":
            self.dirName = dirStr
            self.tilConstFiles = tilLoader.TilConstFiles(dirStr)
        #fileSel.filew.complete("*.til")
        fileSel.filew.set_modal(True)
        fileSel.filew.show()

    def debugCb_toggled(self, widget):
        if widget.get_active():
            print "true"
        else:
            print "false"
        self.debug = widget.get_active()

    def SetBtnsAndLabels(self):
        if self.helloFlag:
            self.helloLabel = "Hello, World!"
        else:
            self.helloLabel = "World, hello there!"

    def StartDisplay(self):
        #fgcolor = gtk.gdk.color_parse("black")
        #widget.modify_fg(state, fgcolor)
        bgcolor = gtk.gdk.color_parse("cyan")
        state = gtk.STATE_NORMAL
        widget = self.wTree.get_widget("dirEventBox")
        widget.modify_bg(state, bgcolor)
        widget = self.wTree.get_widget("fileNameEventBox")
        widget.modify_bg(state, bgcolor)
        widget = self.wTree.get_widget("keyEventBox")
        widget.modify_bg(state, bgcolor)
        widget = self.wTree.get_widget("leftEventBox")
        widget.modify_bg(state, bgcolor)
        #color = gtk.gdk.Color(0,50,50)
        fullFileName = ""
        self.cfgFileName = "TilLoader.txt"
        if os.path.exists(self.cfgFileName):
            f = file(self.cfgFileName, 'r');
            try:
                fullFileName = f.readline()
            finally:
                f.close()
        filePair = os.path.split(fullFileName)
        widget = self.wTree.get_widget("dirLabel")
        widget.set_text(filePair[0])
        self.dirName = filePair[0]
        self.tilConstFiles = tilLoader.TilConstFiles(self.dirName)
        widget = self.wTree.get_widget("fileNameLabel")
        widget.set_text(filePair[1])
        widget = self.wTree.get_widget("tilTextView")
        fontDesc = pango.FontDescription("Monospace Not-Rotated 10")
        widget.modify_font(fontDesc)
        self.debug = False

if __name__ == "__main__":
    hwg = HellowWorldGTK()
    gtk.main()

